__all__ = [
    'AdminCmd',
    'CreateComplexModelCmd',
    'CreateSimpleModelCmd',
    'EnableBashCompleteCmd',
]

import os
import sys

from argparse import RawDescriptionHelpFormatter
from platform import system
from subprocess import (
    CalledProcessError,
    run
)

from ..utils.path import as_path

from .base import OasisBaseCommand
from .inputs import InputValues


class EnableBashCompleteCmd(OasisBaseCommand):
    """
    Adds required command to `.bashrc` Linux or .bash_profile for mac
    so that Command autocomplete works for oasislmf CLI
    """
    formatter_class = RawDescriptionHelpFormatter

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        parser.add_argument(
            '-p', '--bash-conf-file', default=None, required=False,
            help='Cookiecutter JSON file path with all options provided in the file'
        )

    def action(self, args):
        # read in user set target install
        inputs = InputValues(args)

        bash_output_fp = as_path(
            inputs.get('bash_conf_file'),
            'Complex lookup config JSON file path', preexists=False
        )

        # select default bashrc if not set
        if not bash_output_fp:
            default_file = '.bash_profile' if system == 'Darwin' else '.bashrc'
            bash_output_fp = os.path.join(
                os.path.expanduser('~'),
                default_file
            )

        # Prompt user, and then install
        msg_user = 'Running this will append a command to the following file:\n'
        if inputs.confirm_action("{} {}".format(msg_user, bash_output_fp)):
            self.install_autocomplete(bash_output_fp)

    def install_autocomplete(self, target_file=None):
        msg_success = 'Auto-Complete installed.'
        msg_failed = 'install failed'
        msg_installed = 'Auto-Complete feature is already enabled.'
        msg_reload_bash = '\n To activate reload bash by running: \n     source {}'.format(target_file)
        cmd_header = '# Added by OasisLMF\n'
        cmd_autocomplete = 'complete -C completer_oasislmf oasislmf\n'

        try:
            if os.path.isfile(target_file):
                # Check command is in file
                with open(target_file, "r") as rc:
                    if cmd_autocomplete in rc.read():
                        self.logger.info(msg_installed)
                        self.logger.info(msg_reload_bash)
                        sys.exit(0)
            else:
                # create new file at set location
                basedir = os.path.dirname(target_file)
                if not os.path.isdir(basedir):
                    os.makedirs(basedir)

            # Add complete command
            with open(target_file, "a") as rc:
                rc.write(cmd_header)
                rc.write(cmd_autocomplete)
                self.logger.info(msg_success)
                self.logger.info(msg_reload_bash)
        except Exception as e:
            self.logger.error('{}: {}'.format(msg_failed, e))


class CreateSimpleModelCmd(OasisBaseCommand):
    """
    Creates a local Git repository for a "simple model" (using the
    ``cookiecutter`` package) on a "simple model" repository template
    on GitHub)
    """
    formatter_class = RawDescriptionHelpFormatter

    cookiecutter_template_uri = 'git+ssh://git@github.com/OasisLMF/CookiecutterOasisSimpleModel'

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument(
            '-p', '--preset-cookiecutter-json', default=None, required=False, help='Cookiecutter JSON file path with all options provided in the file'
        )
        parser.add_argument(
            '-i', '--no-input', default=None, help='Do not prompt for parameters and only use cookiecutter.json file content', action='store_true'
        )
        parser.add_argument(
            '-r', '--replay', default=None, help='Do not prompt for parameters and only use information entered previously', action='store_true'
        )
        parser.add_argument(
            '-f', '--overwrite-if-exists', default=None, help='Overwrite the contents of any preexisting project directory of the same name', action='store_true'
        )
        parser.add_argument(
            '-o', '--output-dir', default=None, required=False, help='Where to generate the project'
        )
        parser.add_argument(
            '-v', '--cookiecutter-version', default=None, required=False, help='Cookiecutter version', action='store_true'
        )

    def action(self, args):
        """
        Command action

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)

        preset_cookiecutter_json = as_path(
            inputs.get('preset_cookiecutter_json', required=False, is_path=True), label='Preset cookiecutter JSON', is_dir=False, preexists=True
        )

        no_prompt = inputs.get('no_input', default=False, required=False)

        replay = inputs.get('replay', default=False, required=False)

        overwrite = inputs.get('overwrite_if_exists', default=True, required=False)

        pkg_dir = os.path.join(os.path.dirname(__file__), os.pardir)

        target_dir = as_path(
            inputs.get('output_dir', default=os.path.join(pkg_dir, 'cookiecutter-run'), required=False, is_path=True), label='Target directory', is_dir=True, preexists=False
        )

        verbose = inputs.get('verbose', default=False, required=False)

        cookiecutter_version = inputs.get('cookiecutter_version', default=False, required=False)

        def run_cmd(cmd_str):
            run(cmd_str.split(), check=True)

        cmd_str = 'cookiecutter'

        if not cookiecutter_version:
            cmd_str += ''.join([
                (' --no-input' if no_prompt or preset_cookiecutter_json else ''),
                (' --replay' if replay else ''),
                (' --overwrite-if-exists' if overwrite else ''),
                (' --output-dir {}'.format(target_dir) if target_dir else ''),
                (' --verbose ' if verbose else ' '),
                self.cookiecutter_template_uri
            ])
        else:
            cmd_str += ' -V'

        self.logger.info('\nRunning cookiecutter command: {}\n'.format(cmd_str))

        try:
            run_cmd(cmd_str)
        except CalledProcessError as e:
            self.logger.error(e)


class CreateComplexModelCmd(OasisBaseCommand):
    """
    Creates a local Git repository for a "complex model" (using the
    ``cookiecutter`` package) on a "simple model" repository template
    on GitHub)
    """
    formatter_class = RawDescriptionHelpFormatter

    cookiecutter_template_uri = 'git+ssh://git@github.com/OasisLMF/CookiecutterOasisComplexModel'

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super(self.__class__, self).add_args(parser)

        parser.add_argument(
            '-p', '--preset-cookiecutter-json', default=None, required=False, help='Cookiecutter JSON file path with all options provided in the file'
        )
        parser.add_argument(
            '-i', '--no-input', default=None, help='Do not prompt for parameters and only use cookiecutter.json file content', action='store_true'
        )
        parser.add_argument(
            '-r', '--replay', default=None, help='Do not prompt for parameters and only use information entered previously', action='store_true'
        )
        parser.add_argument(
            '-f', '--overwrite-if-exists', default=None, help='Overwrite the contents of any preexisting project directory of the same name', action='store_true'
        )
        parser.add_argument(
            '-o', '--output-dir', default=None, required=False, help='Where to generate the project'
        )
        parser.add_argument(
            '-v', '--cookiecutter-version', default=None, required=False, help='Cookiecutter version', action='store_true'
        )

    def action(self, args):
        """
        Command action

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)

        preset_cookiecutter_json = as_path(
            inputs.get('preset_cookiecutter_json', required=False, is_path=True), label='Preset cookiecutter JSON', is_dir=False, preexists=True
        )

        no_prompt = inputs.get('no_input', default=False, required=False)

        replay = inputs.get('replay', default=False, required=False)

        overwrite = inputs.get('overwrite_if_exists', default=True, required=False)

        pkg_dir = os.path.join(os.path.dirname(__file__), os.pardir)

        target_dir = as_path(
            inputs.get('output_dir', default=os.path.join(pkg_dir, 'cookiecutter-run'), required=False, is_path=True), label='Target directory', is_dir=True, preexists=False
        )

        verbose = inputs.get('verbose', default=False, required=False)

        cookiecutter_version = inputs.get('cookiecutter_version', default=False, required=False)

        def run_cmd(cmd_str):
            run(cmd_str.split(), check=True)

        cmd_str = 'cookiecutter'

        if not cookiecutter_version:
            cmd_str += ''.join([
                (' --no-input' if no_prompt or preset_cookiecutter_json else ''),
                (' --replay' if replay else ''),
                (' --overwrite-if-exists' if overwrite else ''),
                (' --output-dir {}'.format(target_dir) if target_dir else ''),
                (' --verbose ' if verbose else ' '),
                self.cookiecutter_template_uri
            ])
        else:
            cmd_str += ' -V'

        self.logger.info('\nRunning cookiecutter command: {}\n'.format(cmd_str))

        try:
            run_cmd(cmd_str)
        except CalledProcessError as e:
            self.logger.error(e)


class AdminCmd(OasisBaseCommand):
    """
    Admin subcommands::

        * creates a local Git repository for a "simple model" (using the
          ``cookiecutter`` package) on a "simple model" repository template
          on GitHub)
        * creates a local Git repository for a "complex model" (using the
          ``cookiecutter`` package) on a "complex model" repository template
          on GitHub)
    """
    sub_commands = {
        'create-simple-model': CreateSimpleModelCmd,
        'create-complex-model': CreateComplexModelCmd,
        'enable-bash-complete': EnableBashCompleteCmd
    }



import oasislmf.cli.base
import tempfile
import os

from argparsetree import BaseCommand

from ..model_testing.validation import csv_validity_test

from ..utils.path import as_path

from ..utils.exceptions import OasisException
from .base import OasisBaseCommand
from .inputs import InputValues

from ..manager import OasisManager as om

# ! Handle multiple tests


class ModelValidationCmd(OasisBaseCommand):
    """
    Checks the validity of a set of model data.
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """

        super(self.__class__, self).add_args(parser)

        parser.add_argument(
            '-d', '--model-data-path',
            default=None, help='Directory containing additional user-supplied model data files')

    def action(self, args):
        """
        Performs validity checks on model data csv files using ktools
        executables.

        :param args: The arguments from the command line
        :type args: Namespace
        """

        inputs = InputValues(args)

        model_data_fp = as_path(
            inputs.get('model_data_path', required=True, is_path=True), 'Model data path', is_dir=True)

        csv_validity_test(model_data_fp)


class FmValidationCmd(OasisBaseCommand):
    """
    Runs a set of FM tests.
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """

        super(self.__class__, self).add_args(parser)

        parser.add_argument(
            '-c', '--test-case-name',
            type=str, default=None, required=False,
            help='Test case name - runs a specific test in the test directory. Otherwise run all tests.'
        )
        parser.add_argument(
            '-l', '--list-tests',
            action='store_true',
            help='List the valid test cases in the test directory rather than running'
        )
        parser.add_argument(
            '-t', '--test-dir',
            type=str, default=None, required=True,
            help='Test directory - should contain test directories containing OED files and expected results'
        )
        parser.add_argument(
            '-r', '--run-dir',
            type=str, default=None, required=False,
            help='Run directory - where files should be generated. If not sst, no files will be saved.'
        )

    def action(self, args):
        """
        Runs a set of FM tests.

        :param args: The arguments from the command line
        :type args: Namespace
        """

        inputs = InputValues(args)

        call_dir = os.getcwd()

        do_list_tests = inputs.get('list_tests')

        test_case_name = inputs.get('test_case_name')
        run_all_tests = test_case_name is None

        test_dir = as_path(
            inputs.get(
                'test_dir', default=call_dir, is_path=True),
            'Source files directory', is_dir=True, preexists=True)

        test_case_names = []
        if run_all_tests:
            for test_case in os.listdir(path=test_dir):
                if os.path.exists(
                    os.path.join(test_dir, test_case, "expected")
                ):
                    test_case_names.append(test_case)
        else:
            test_case_names.append(test_case_name)

        test_case_names.sort()

        status = 0

        if do_list_tests:
            for test_case_name in test_case_names:
                print(test_case_name)
            exit(0)

        failed_tests = []

        for test_case_name in test_case_names:

            test_case_dir = os.path.join(test_dir, test_case_name)
            if not os.path.exists(test_case_dir):
                raise OasisException(f"Test directory does not exist: {test_case_name}")

            run_dir = as_path(inputs.get('run_dir', is_path=True), 'Run directory', is_dir=True, preexists=False)
            if run_dir is None:
                with tempfile.TemporaryDirectory() as tmp_run_dir:
                    test_result = om().run_fm_test(test_case_dir, tmp_run_dir)
            else:
                run_dir = os.path.join(run_dir, test_case_name)
                test_result = om().run_fm_test(test_case_dir, run_dir)

            if not test_result:
                failed_tests.append(test_case_name)
                if status == 0:
                    status = 1

        if len(failed_tests) == 0:
            self.logger.info("All tests passed")
        else:
            self.logger.info("{} test failed: ".format(len(failed_tests)))
            [self.logger.info(n) for n in failed_tests]

        exit(status)


class TestCmd(BaseCommand):
    """
    Test models and keys servers
    """

    sub_commands = {
        'model-validation': ModelValidationCmd,
        'fm': FmValidationCmd
    }
